from dataclasses import dataclass
from typing import List

import pandas as pd


@dataclass()
class ProfileMetaData:
    """
    Single instance of Metadata associated with a profile
    """
    site_name: str
    date_time: pd.Timestamp
    latitude: float
    longitude: float
    utm_epsg: str = None  # the EPSG for the utm zone
    campaign_name: str = None
    flags: str = None
    comments: str = None
    observers: List[str] = None



from insitupy.variables.base_metadata_variables import BaseMetadataVariables
from insitupy.variables.base_variables import MeasurementDescription


class SnowExMetadataVariables(BaseMetadataVariables):
    """
    Extend the metadata variables to include more mappings
    for snowex
    """
    TOTAL_DEPTH = MeasurementDescription(
        'total_depth', "Total depth of measurement",
        ['total_snow_depth'], auto_remap=True
    )
    LATITUDE = MeasurementDescription(
        'latitude', "Latitude",
        ['lat', 'latitude'], auto_remap=True
    )
    LONGITUDE = MeasurementDescription(
        'longitude', "Longitude",
        ['long', 'lon', 'longitude'], auto_remap=True
    )
    NORTHING = MeasurementDescription(
        'northing', "UTM Northing",
        ['northing', 'utm_wgs84_northing'], auto_remap=True
    )
    EASTING = MeasurementDescription(
        'easting', "UTM Easting",
        ['easting', 'utm_wgs84_easting'], auto_remap=True
    )
    UTM_ZONE = MeasurementDescription(
        'utm_zone', "UTM Zone",
        ['utmzone', 'utm_zone'], auto_remap=True
    )
    ELEVATION = MeasurementDescription(
        'elevation', "Elevation",
        ['elev_m', 'elevation', 'elevationwgs84'], auto_remap=True
    )



from insitupy.variables.base_variables import ExtendableVariables, \
    MeasurementDescription


class BaseMetadataVariables(ExtendableVariables):
    """
    Variables expected to be found in file headers
    These are variables associated with larger dataset
    """
    DATETIME = MeasurementDescription(
        'datetime', "Combined date and time",
        ["Date/Local Standard Time", "date/local_standard_time", "datetime",
         "date&time", "date/time", "date/local_time"],
        True
    )
    DATE = MeasurementDescription(
        'date', "Measurement Date (only date column)",
        ['date_dd_mmm_yy', 'date']
    )
    TIME = MeasurementDescription(
        'time', "Measurement time",
        ['time_gmt', 'time']
    )
    TIME_BOUND_PIT = MeasurementDescription(
        "Time start/end",
        "Time of first or last pit measurement",
        ["Time start/end", "time_start/end"]
    )
    SITE_NAME = MeasurementDescription(
        "site_name", "Name of campaign site",
        ['location'], True
    )
    SITE_ID = MeasurementDescription(
        "site_id", "ID within campaign site",
        ['site'], True
    )
    PIT_ID = MeasurementDescription(
        "pit_id", "ID of snow pit",
        ['pitid'], True
    )



from .metadata_variables import SnowExMetadataVariables
from .primary_variables import SnowExPrimaryVariables
from .snowex_campaign import SnowExProfileData, SnowExMetadataParser

__all__ = [
    "SnowExProfileData", "SnowExMetadataParser", "SnowExMetadataVariables",
    "SnowExPrimaryVariables"
]



import pandas as pd
import pytest

from insitupy.campaigns.snowex import SnowExMetadataParser


@pytest.mark.parametrize(
    "fname", [
        "SNEX20_TS_SP_20200427_0845_COERAP_data_density_v01.csv",
        "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
        "SNEX20_TS_SP_20200427_0845_COERAP_data_temperature_v01.csv"
    ]
)
class TestSnowexPitMetadata:
    """
    Test that we can consistently read metadata across
    multiple pit measurements
    """

    @pytest.fixture
    def metadata_info(self, fname, data_path):
        # This is the parser object
        obj = SnowExMetadataParser(
            data_path.joinpath(fname), "US/Mountain",
            allow_map_failures=True
        )
        metadata, columns, column_mapping, header_pos = obj.parse()
        return metadata, columns, column_mapping, header_pos

    @pytest.fixture
    def metadata(self, metadata_info):
        return metadata_info[0]

    @pytest.fixture
    def columns(self, metadata_info):
        return metadata_info[1]

    @pytest.fixture
    def header_pos(self, metadata_info):
        return metadata_info[3]

    # fname needs to be in test since it is a class
    # level parameterization
    def test_id(self, metadata, fname):
        assert metadata.site_name == "COERAP_20200427_0845"

    def test_date_time(self, metadata, fname):
        # converted from mountain time
        assert metadata.date_time == pd.to_datetime(
            "2020-04-27T14:45:00+0000"
        )

    def test_latitude(self, metadata, fname):
        assert metadata.latitude == 38.92524

    def test_longitude(self, metadata, fname):
        assert metadata.longitude == -106.97112

    def test_utm_epsg(self, metadata, fname):
        assert metadata.utm_epsg == '26913'

    def test_site_name(self, metadata, fname):
        assert metadata.campaign_name == "East River"

    def test_flags(self, metadata, fname):
        assert metadata.flags is None

    def test_header_position(self, header_pos, fname):
        assert header_pos == 10


@pytest.mark.parametrize(
    "fname, expected_cols", [
        ("SNEX20_TS_SP_20200427_0845_COERAP_data_density_v01.csv",
         ['depth', 'bottom_depth', 'density_a', 'density_b', 'density_c']),
        ("SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
         ['depth', 'bottom_depth', 'density', 'permittivity_a',
          'permittivity_b', 'lwc_vol_a', 'lwc_vol_b']),
        ("SNEX20_TS_SP_20200427_0845_COERAP_data_temperature_v01.csv",
         ['depth', 'snow_temperature'])
    ]
)
def test_columns(fname, expected_cols, data_path):
    """
    Test the columns we expect to pass back from the file
    """
    obj = SnowExMetadataParser(
        data_path.joinpath(fname), "US/Mountain",
        allow_map_failures=True
    )
    metadata, columns, column_mapping, header_pos = obj.parse()
    assert columns == expected_cols



from insitupy.variables.base_variables import ExtendableVariables, \
    MeasurementDescription


class BasePrimaryVariables(ExtendableVariables):
    """
    Variables expected to be found in COLUMNS
    These are variables associated with the line-by-line
    measurements
    """
    # TODO optional multisample list to check

    SWE = MeasurementDescription(
        "SWE", "Snow Water Equivalent",
        ["swe_mm", "swe"], auto_remap=True
    )
    DEPTH = MeasurementDescription(
        "depth", "top or center depth of measurement",
        [
            "depth", "top", "sample_top_height", "hs",
            "depth_m", 'snowdepthfilter(m)', 'snowdepthfilter',
            'height'
        ], auto_remap=True
    )
    BOTTOM_DEPTH = MeasurementDescription(
        "bottom_depth", "Lower edge of measurement",
        ["bottom", "bottom_depth"], auto_remap=True
    )
    DENSITY_A = MeasurementDescription(
        "density", "measured snow density",
        ["density_a"], auto_remap=False, match_on_code=False
    )
    DENSITY_B = MeasurementDescription(
        "density", "measured snow density",
        ["density_b"], auto_remap=False, match_on_code=False
    )
    DENSITY_C = MeasurementDescription(
        "density", "measured snow density",
        ["density_c"], auto_remap=False, match_on_code=False
    )
    DENSITY = MeasurementDescription(
        "density", "measured snow density",
        [
            "density", "avg_density",
            "avgdensity", 'density_mean'
        ], auto_remap=True
    )
    LAYER_THICKNESS = MeasurementDescription(
        "layer_thickness", "thickness of layer", auto_remap=True
    )
    SNOW_TEMPERATURE = MeasurementDescription(
        "snow_temperature", "Snowpack Temperature",
        ["temperature", "temperature_deg_c"], auto_remap=True
    )
    LWC_A = MeasurementDescription(
        "liquid_water_content", "Liquid water content",
        ["lwc_vol_a"], auto_remap=False, match_on_code=False
    )
    LWC_B = MeasurementDescription(
        "liquid_water_content", "Liquid water content",
        ["lwc_vol_b"], auto_remap=False, match_on_code=False
    )
    LWC = MeasurementDescription(
        "liquid_water_content", "Liquid water content",
        ["lwc", "lwc_vol"], auto_remap=True
    )
    PERMITTIVITY_A = MeasurementDescription(
        "permittivity", "Permittivity",
        ["permittivity_a", 'dielectric_constant_a'],
        auto_remap=False, match_on_code=False
    )
    PERMITTIVITY_B = MeasurementDescription(
        "permittivity", "Permittivity",
        ["permittivity_b", 'dielectric_constant_b'],
        auto_remap=False, match_on_code=False
    )
    PERMITTIVITY = MeasurementDescription(
        "permittivity", "Permittivity",
        ["permittivity", 'dielectric_constant'], True
    )
    GRAIN_SIZE = MeasurementDescription(
        "grain_size", "Grain Size",
        ["grain_size"], auto_remap=True
    )
    GRAIN_TYPE = MeasurementDescription(
        "grain_type", "Grain Type",
        ["grain_type"], auto_remap=True
    )
    HAND_HARDNESS = MeasurementDescription(
        "hand_hardness", "Hand Hardness",
        ["hand_hardness"], auto_remap=True
    )
    MANUAL_WETNESS = MeasurementDescription(
        "manual_wetness", "Manual Wetness",
        ["manual_wetness"], auto_remap=True
    )



import logging

import numpy as np
import pandas as pd
import geopandas as gpd

from insitupy.io.metadata import MetaDataParser
from .metadata import ProfileMetaData
from insitupy.variables import MeasurementDescription, ExtendableVariables


LOG = logging.getLogger(__name__)


class ProfileData:
    """
    This would be one pit, SMP profile, etc
    Unique date, location, variable
    """
    META_PARSER = MetaDataParser

    def __init__(
        self, input_df: pd.DataFrame, metadata: ProfileMetaData,
        variable: MeasurementDescription,
        original_file=None,
        units_map=None,
        allow_map_failure=False
    ):
        """
        Take df of layered data (SMP, pit, etc)
        Args:
            input_df: dataframe of data
                Should include depth and optional bottom depth
                Should include sample or sample_a, sample_b, etc
            metadata: ProfileMetaData object
            variable: description of variable
            original_file: optional track original file
            units_map: optional dictionary of column name to unit
            allow_map_failures: if a mapping fails, warn us and use the
                original string (default False)

        """
        self._units_map = units_map
        self._original_file = original_file
        self._depth_layer = self.depth_columns()[0]
        self._lower_depth_layer = self.depth_columns()[1]
        self._metadata = metadata
        self._allow_map_failure = allow_map_failure
        self.variable: MeasurementDescription = variable
        # mapping of column name to measurement type
        self._column_mappings = {}
        # List of measurements to keep
        self._measurements_to_keep = (
            self.shared_column_options() + [self.variable]
        )

        self._id = metadata.site_name
        self._dt = metadata.date_time

        self._sample_column = None
        # This will populate the column mapping
        # and filter to the desired measurement column
        if input_df.empty:
            LOG.warning("DF is empty")
            self._df = input_df
        else:
            self._df = self._format_df(input_df)

        columns = self._df.columns.values
        if len(columns) > 0 and self._depth_layer.code not in columns:
            raise ValueError(f"Expected {self._depth_layer} in columns")

        # List of columns that are not the desired variable
        _non_measure_columns = [
            self._depth_layer.code, self._lower_depth_layer.code,
            "datetime",
            "geometry"
        ]
        self._non_measure_columns = [
            c for c in _non_measure_columns if c in columns
        ]

        # describe the data a bit
        self._has_layers = self._lower_depth_layer.code in columns
        # Extend the df info
        self._add_thickness_to_df()

    @classmethod
    def depth_columns(cls):
        return [
            cls.META_PARSER.PRIMARY_VARIABLES_CLASS.DEPTH,
            cls.META_PARSER.PRIMARY_VARIABLES_CLASS.BOTTOM_DEPTH
        ]

    @classmethod
    def shared_column_options(cls):
        return cls.depth_columns()

    def _set_column_mappings(self, input_df):
        # Get rid of columns we don't want and populate column mapping
        columns = input_df.columns.values
        for c in columns:
            # Find the variable associated with each column
            # and store a map
            cn, cm = self.META_PARSER.PRIMARY_VARIABLES_CLASS.from_mapping(
                c, allow_failure=self._allow_map_failure
            )
            # join with existing mappings
            self._column_mappings = {**cm, **self._column_mappings}

    def _check_sample_columns(self, input_df):
        columns = input_df.columns.values
        _sample_columns = [
            c for c in columns if self._column_mappings.get(c) == self.variable
        ]
        if len(_sample_columns) == 0:
            raise ValueError(
                f"No sample columns in {columns}. This is likely because"
                f" variable {self.variable} is not in columns {input_df.columns}"
            )
        elif len(_sample_columns) > 1:
            LOG.warning("Only one sample column allowed, keeping the first match")
            _sample_columns = [_sample_columns[0]]
        _sample_column = _sample_columns[0]
        # Rename the variable column to the variable code
        # This covers the scenario where we did not auto-remap the
        # variable on original parsing of the column names because
        # it was a multi-sample variable
        # Columns related to the variable
        sample_column_type = self._column_mappings[_sample_column]
        if sample_column_type != self.variable:
            raise ValueError(
                f"{sample_column_type} and {self.variable} are not the same"
            )
        input_df.rename(
            columns={_sample_column: sample_column_type.code},
            inplace=True
        )
        self._sample_column = sample_column_type.code
        return input_df

    def _format_df(self, input_df):
        """
        Format the incoming df with the column headers and other info we want
        This will filter to a single measurement as well as the expected
        shared columns like depth
        """
        columns = input_df.columns.values
        self._set_column_mappings(input_df)

        columns_to_keep = [
            c for c in columns
            if self._column_mappings[c] in self._measurements_to_keep
        ]
        df = input_df.loc[:, columns_to_keep]
        # Verify the sample column exists and rename to variable
        df = self._check_sample_columns(df)

        n_entries = len(df)
        df["datetime"] = [self._dt] * n_entries

        # parse the location
        lat, lon = self.latlon
        location = gpd.points_from_xy(
            [lon] * n_entries, [lat] * n_entries
        )

        df = gpd.GeoDataFrame(
            df, geometry=location
        ).set_crs("EPSG:4326")
        df = df.replace(-9999, np.NaN)

        return df

    def _add_thickness_to_df(self):
        # set the thickness of the layer
        if self._has_layers:
            self._df[self.META_PARSER.PRIMARY_VARIABLES_CLASS.LAYER_THICKNESS.code] = (
                self._df[self._depth_layer.code] - self._df[
                    self._lower_depth_layer.code
                ]
            )

    @property
    def metadata(self):
        return self._metadata

    @property
    def units_map(self):
        return self._units_map

    @property
    def latlon(self):
        # return location metadata
        return self._metadata.latitude, self._metadata.longitude

    @property
    def df(self):
        return self._df

    @property
    def sum(self):
        # get bulk value
        if not self._has_layers:
            # could we assume equidistant spacing and do a mean?
            raise RuntimeError("Cannot compute for no layers")

        # this should work for multi or not multi sample
        profile = self._df.loc[:, self._sample_column]
        self._df["mean"] = profile
        # TODO: sum up with depth change
        # TODO: could we use the weighted mean * the total depth?
        # TODO: units
        pass

    @property
    def mean(self):
        profile = self._df.loc[:, self._sample_column]
        if pd.isna(profile).all():
            return np.nan
        if self._has_layers:
            # height weighted mean for these layers
            thickness = self._df[
                self.META_PARSER.PRIMARY_VARIABLES_CLASS.LAYER_THICKNESS.code
            ]
            # this works for a weighted mean, but is not assumed to be
            # the total thickness of the snowpack
            thickness_total = thickness.sum()
            weighted_mean = (
                profile * thickness / thickness_total
            ).sum()
            value = weighted_mean
        else:
            value = np.mean(profile)

        return value

    @property
    def total_depth(self):
        profile = self._df.loc[:, self._depth_layer.code].values
        return np.nanmax(profile)

    def get_profile(self, snow_datum="ground"):
        # TODO: snow datum is ground or snow
        # get profile of values
        profile_average = self._df.loc[:, self._sample_column]
        df = self._df.copy()
        df[self.variable.code] = profile_average
        columns_of_interest = [*self._non_measure_columns, self.variable.code]
        return df.loc[:, columns_of_interest]

    @staticmethod
    def read_csv_dataframe(profile_filename, columns, header_position):
        """
        Read in a profile file. Managing the number of lines to skip and
        adjusting column names

        Args:
            profile_filename: Filename containing a manually measured
                             profile
            columns: list of columns to use in dataframe
            header_position: skiprows for pd.read_csv
        Returns:
            df: pd.dataframe contain csv data with desired column names
        """
        raise NotImplementedError("Not implemented")

    @classmethod
    def from_csv(
        self, fname, variable: ExtendableVariables, allow_map_failures=False
    ):
        raise NotImplementedError("Not implemented")


def standardize_depth(depths, desired_format='snow_height', is_smp=False):
    """
    Data that is a function of depth comes in 2 formats. Sometimes 0 is
    the snow surface, sometimes 0 is the ground. This function standardizes it
    for each profile. desired_format can be:

        snow_height: Zero at the bottom of the data.
        surface_datum: Zero at the top of the data and uses negative depths
                       (easier for plotting)

    Args:
        depths: Pandas series of depths in either format
        desired_format: string indicating which format the data is in
        is_smp: Boolean indicating which data this is, if smp then the data is
                surface_datum but with positive depths
   Returns:
        new:
    """
    max_depth = depths.max()
    min_depth = depths.min()

    new = depths.copy()

    # How is the depth ordered
    # max_depth_at_top = depths.iloc[0] > depths.iloc[-1]

    # Is the data in surface_datum already
    bottom_is_negative = depths.iloc[-1] < 0

    if desired_format == 'snow_height':

        if is_smp:
            LOG.info('Converting SMP depths to snow height format.')
            new = (depths - max_depth).abs()

        elif bottom_is_negative:
            LOG.info('Converting depths in surface datum to snow height format.')

            new = (depths + abs(min_depth))

    elif desired_format == 'surface_datum':
        if is_smp:
            LOG.info('Converting SMP depths to surface datum format.')
            new = depths.mul(-1)

        elif not bottom_is_negative:
            LOG.info('Converting depths in snow height to surface datum format.')
            new = depths - max_depth

    else:
        raise ValueError(
            f'{desired_format} is an invalid depth format! Options are:'
            f' {["snow_height", "surface_datum"]}'
        )

    return new



"""
Point data from select manual measurement campaigns
"""
import logging
from typing import List

import pandas as pd
from insitupy.io.metadata import MetaDataParser, ProfileMetaData
from insitupy.profiles.base import ProfileData
from insitupy.variables import MeasurementDescription


SOURCES = [
    "https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_SSA.001/",
    "https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_BSU_GPR.001/",
    "https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_GM_SP.001/",
    "https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_SMP.001/",
    "https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_SD.001/",
    ("https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_GM_CSU_GPR.001/2020.02.06/"
     "SNEX20_GM_CSU_GPR_1GHz_v01.csv"),
    ("https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_UNM_GPR.001/2020.01.28/"
     "SNEX20_UNM_GPR.csv"),
    ("https://n5eil01u.ecs.nsidc.org/SNOWEX/SNEX20_SD_TLI.001/2019.09.29/"
     "SNEX20_SD_TLI_clean.csv"),
]


LOG = logging.getLogger(__name__)


class ProfileDataCollection:
    """
    This could be a collection of profiles
    """
    META_PARSER = MetaDataParser
    PROFILE_DATA_CLASS = ProfileData

    def __init__(self, profiles: List[ProfileData], metadata: ProfileMetaData):
        self._profiles = profiles
        self._metadata = metadata

    @property
    def SWE(self):
        """
        Takes all layers for each unique date, location and returns point swe
        geodataframe
        We can iterate though all ProfileData in this class and get SWE DF for each
        """
        # find all points with variable == density and calc SWE
        pass

    def get_mean(self, variable: MeasurementDescription):
        raise NotImplementedError()

    def get_sum(self, variable: MeasurementDescription):
        raise NotImplementedError()

    def get_profile(self, variable: MeasurementDescription):
        raise NotImplementedError()

    @property
    def metadata(self) -> ProfileMetaData:
        return self._metadata

    @property
    def profiles(self) -> List[ProfileData]:
        return self._profiles

    @classmethod
    def _read_csv(
        cls, fname, columns, column_mapping, header_pos,
        metadata: ProfileMetaData, units_map
    ) -> List[ProfileData]:
        """
        Args:
            fname: path to csv
            columns: columns for dataframe
            column_mapping: mapping of column name to variable description
            header_pos: skiprows for pd.read_csv
            metadata: metadata for each object
            units_map: map of column name to unit

        Returns:
            a list of ProfileData objects

        """
        result = []
        if columns is None and header_pos is None:
            LOG.warning(f"File {fname} is empty of rows")
            df = pd.DataFrame()
        else:
            df = cls.PROFILE_DATA_CLASS.read_csv_dataframe(
                fname, columns, header_pos
            )
        # add comments in here for shared columns
        shared_column_options = cls.PROFILE_DATA_CLASS.shared_column_options()
        shared_columns = [
            c for c, v in column_mapping.items()
            if v in shared_column_options
        ]
        variable_columns = [
            c for c in column_mapping.keys() if c not in shared_columns
        ]

        # Create an object for each measurement
        for column in variable_columns:
            target_df = df.loc[:, shared_columns + [column]]
            result.append(cls.PROFILE_DATA_CLASS(
                target_df, metadata,
                column_mapping[column],  # variable is a MeasurementDescription
                original_file=fname,
                units_map=units_map,
            ))
        if not result and df.empty:
            # Add one profile if this is empty so we can
            # keep the metadata
            result = [
                cls.PROFILE_DATA_CLASS(
                    df, metadata,
                    None,
                    # variable is a MeasurementDescription
                    original_file=fname,
                    units_map=units_map
                )
            ]

        return result

    @classmethod
    def from_csv(
        cls, fname, timezone="US/Mountain", header_sep=",", site_id=None,
        campaign_name=None, allow_map_failure=False
    ):
        """
        Find all profiles in a single csv file
        Args:
            fname: path to file
            timezone: expected timezone in file
            header_sep: header sep in the file
            site_id: Site id override for the metadata
            campaign_name: Campaign.name override for the metadata
            allow_map_failure: allow metadata and column unknowns

        Returns:
            This class with a collection of profiles and metadata
        """
        # TODO: timezone here (mapped from site?)
        # parse mlutiple files and create an iterable of ProfileData
        meta_parser = cls.META_PARSER(
            fname, timezone, header_sep=header_sep, _id=site_id,
            campaign_name=campaign_name, allow_map_failures=allow_map_failure
        )
        # Parse the metadata and column info
        metadata, columns, columns_map, header_pos = meta_parser.parse()
        # read in the actual data
        profiles = cls._read_csv(
            fname, columns, columns_map, header_pos, metadata,
            meta_parser.units_map
        )

        # ignore profiles with the name 'ignore'
        profiles = [
            p for p in profiles if
            # Keep the profile if it is None because we need the metadata
            (p.variable is None or p.variable.code != "ignore")
        ]

        return cls(profiles, metadata)



import pytest
import numpy as np

from insitupy.variables import BasePrimaryVariables
from insitupy.campaigns import ProfileDataCollection
from insitupy.campaigns.snowex import SnowExMetadataParser, SnowExProfileData


class DataCollectionExtended(ProfileDataCollection):
    META_PARSER = SnowExMetadataParser
    PROFILE_DATA_CLASS = SnowExProfileData


@pytest.mark.parametrize(
    "fname, expected_variables, expected_means", [
        (
            "SNEX20_TS_SP_20200427_0845_COERAP_data_temperature_v01.csv",
            [BasePrimaryVariables.SNOW_TEMPERATURE], [0.0]
        ),
        (
            "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
            [
                BasePrimaryVariables.DENSITY,
                BasePrimaryVariables.PERMITTIVITY_A,
                BasePrimaryVariables.PERMITTIVITY_B,
                BasePrimaryVariables.LWC_A,
                BasePrimaryVariables.LWC_B,
            ], [395.03703703703707, np.nan, np.nan, np.nan, np.nan]
        ),
        (
            "SNEX20_TS_SP_20200427_0845_COERAP_data_density_v01.csv",
            [
                BasePrimaryVariables.DENSITY_A, BasePrimaryVariables.DENSITY_B,
                BasePrimaryVariables.DENSITY_C
            ],
            [397.8888888, 394.5555556, 137.1111111]
        ),
    ]
)
class TestProfileDataCollection:
    @pytest.fixture
    def obj(self, fname, data_path):
        file_path = data_path.joinpath(fname)
        obj = DataCollectionExtended.from_csv(
            file_path, allow_map_failure=True
        )
        return obj

    def test_variables(self, obj, expected_variables, expected_means):
        result_variables = [
            p.variable for p in obj.profiles
        ]
        assert result_variables == expected_variables

    def test_mean(self, obj, expected_variables, expected_means):
        result_means = [
            p.mean for p in obj.profiles
        ]
        np.testing.assert_almost_equal(
            np.array(result_means), np.array(expected_means)
        )



from .base_metadata_variables import BaseMetadataVariables
from .base_primary_variables import BasePrimaryVariables
from .base_variables import MeasurementDescription, ExtendableVariables

__all__ = [
    "BasePrimaryVariables", "BaseMetadataVariables", "MeasurementDescription",
    "ExtendableVariables"
]



import inspect
import logging
from dataclasses import dataclass
from typing import List, Tuple, Dict


LOG = logging.getLogger(__name__)


# Similar to MeasurementDescription from metloom
@dataclass(eq=True, frozen=True)
class MeasurementDescription:
    """
    data class for describing a measurement
    """

    code: str = "-1"  # code used within the applicable API
    description: str = None  # description of the sensor
    map_from: List = None  # map to this variable from a list of options
    auto_remap: bool = False  # Auto remap to the column to the code
    match_on_code: bool = True  # Match on the code too
    cast_type = None   # make this float, int, etc


class ExtendableVariables:
    """
    Make a class with the helpful iterator portion of enums, but
    that is fully extendable
    """
    def __init__(self, entries=None):
        self._entries = None
        # allows filtering to a desired list
        self._initial_entries = entries or self._all_variables

    @property
    def entries(self):
        if self._entries is None:
            self._entries = self._get_members()
        return self._entries

    @property
    def _all_variables(self):
        return [m[1] for m in self.entries]

    @property
    def variables(self):
        return [m[1] for m in self.entries if m[1] in self._initial_entries]

    @property
    def names(self):
        return [m[0] for m in self.entries if m[1] in self._initial_entries]

    @classmethod
    def _get_members(cls):
        """
        https://www.geeksforgeeks.org/how-to-get-a-list-of-class-attributes-in-python/
        Important this is a class method or seg faults
        """
        result = []
        # getmembers() returns all the
        # members of an object
        for i in inspect.getmembers(cls):

            # to remove private and protected
            # functions
            if not i[0].startswith('_'):

                # To remove other methods that
                # doesnot start with a underscore
                if not inspect.ismethod(i[1]) and not isinstance(i[1], property):
                    result.append(i)
        return result

    def __iter__(self):
        for item in self.variables:
            yield item

    def __len__(self):
        return len(self.entries)

    @classmethod
    def from_mapping(
        cls, input_name, allow_failure=False
    ) -> Tuple[str, Dict[str, MeasurementDescription]]:
        """
        Get the measurement description from an input name.
        This will use the  MeasurementDescription.map_from list to
        check, in order, which variable we should use

        Args:
            input_name: string input name
            allow_failure: return the original name if we fail

        Returns:
            column name
            column mapping (map of name to MeasurementDescription)
        """
        lower_name = input_name.lower()
        result = None
        # Map column name to variable type
        column_mapping = {}
        for entry in cls():
            # Check if we match directly on the code
            code_match = entry.match_on_code and lower_name == entry.code
            # Check if we match from the list of possible matches
            map_match = entry.map_from and lower_name in entry.map_from
            if code_match or map_match:
                # Remap to code
                if entry.auto_remap:
                    result = entry.code
                else:
                    result = input_name
                # store a map of the column name to the variable description
                column_mapping[result] = entry
                break
        if result is None:
            if allow_failure:
                # We failed to find a mapping, but want to continue
                LOG.warning(f"Could not find mapping for {input_name}")
                result = input_name
                column_mapping[result] = None
            else:
                raise RuntimeError(f"Could not find mapping for {input_name}")
        LOG.debug(
            f"Mapping {result} to {result} (type {column_mapping[result]})"
        )
        return result, column_mapping



import numpy as np
import pytest

from insitupy.campaigns.snowex import SnowExProfileData
from insitupy.variables import BasePrimaryVariables


class TestSnowexPitProfile:
    """
    Test the attributes of the profile
    """

    @pytest.mark.parametrize(
        "fname, variable", [
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
                BasePrimaryVariables.PERMITTIVITY
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_density_v01.csv",
                BasePrimaryVariables.DENSITY
            ),
        ]
    )
    def test_read_fails(self, fname, variable, data_path):
        file_path = data_path.joinpath(fname)
        with pytest.raises(ValueError):
            SnowExProfileData.from_csv(
                file_path, variable, allow_map_failures=True
            )

    @pytest.mark.parametrize(
        "fname, variable", [
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
                BasePrimaryVariables.PERMITTIVITY
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_density_v01.csv",
                BasePrimaryVariables.DENSITY
            ),
        ]
    )
    def test_read_fails_mapping_error(self, fname, variable, data_path):
        file_path = data_path.joinpath(fname)
        with pytest.raises(RuntimeError):
            SnowExProfileData.from_csv(
                file_path, variable, allow_map_failures=False
            )

    @pytest.mark.parametrize(
        "fname, variable, expected", [
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_temperature_v01.csv",
                BasePrimaryVariables.SNOW_TEMPERATURE, 0.0
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
                BasePrimaryVariables.LWC_A, np.nan
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
                BasePrimaryVariables.PERMITTIVITY_A, np.nan
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_density_v01.csv",
                BasePrimaryVariables.DENSITY_A, 397.8888888
            ),
        ]
    )
    def test_mean(self, fname, variable, expected, data_path):
        file_path = data_path.joinpath(fname)
        obj = SnowExProfileData.from_csv(
            file_path, variable, allow_map_failures=True
        )
        result = obj.mean
        if np.isnan(expected):
            assert np.isnan(result)
        else:
            assert result == pytest.approx(expected)

    @pytest.mark.parametrize(
        "fname, variable, expected", [
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_temperature_v01.csv",
                BasePrimaryVariables.SNOW_TEMPERATURE, 95.0
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
                BasePrimaryVariables.LWC_A, 95.0
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_LWC_v01.csv",
                BasePrimaryVariables.PERMITTIVITY_B, 95.0
            ),
            (
                "SNEX20_TS_SP_20200427_0845_COERAP_data_density_v01.csv",
                BasePrimaryVariables.DENSITY_A, 95.0
            ),
        ]
    )
    def test_total_depth(self, fname, variable, expected, data_path):
        file_path = data_path.joinpath(fname)
        obj = SnowExProfileData.from_csv(
            file_path, variable, allow_map_failures=True
        )
        result = obj.total_depth
        assert result == pytest.approx(expected)



import logging

import pandas as pd
from pathlib import Path

from insitupy.io.metadata import MetaDataParser
from . import SnowExMetadataVariables, SnowExPrimaryVariables
from insitupy.variables import MeasurementDescription
from insitupy.profiles.base import ProfileData, standardize_depth


LOG = logging.getLogger(__name__)


class SnowExMetadataParser(MetaDataParser):
    METADATA_VARIABLE_CLASS = SnowExMetadataVariables
    PRIMARY_VARIABLES_CLASS = SnowExPrimaryVariables


class SnowExProfileData(ProfileData):
    META_PARSER = SnowExMetadataParser

    @classmethod
    def from_csv(
        cls, fname, variable: MeasurementDescription, timezone="US/Mountain",
        allow_map_failures=False
    ):
        """
        Args:
            fname: path to file
            variable: variable in the file
            timezone: local timezone for file
            allow_map_failures: allow metadata and column unknowns
        Returns:
            the instantiated class
        """
        meta_parser = cls.META_PARSER(
            fname, timezone, allow_map_failures=allow_map_failures
        )
        # Parse the metadata and column info
        metadata, columns, columns_map, header_pos = meta_parser.parse()
        # read in the actual data
        if columns is None and not columns_map:
            # Use an empty dataframe if the file is empty
            LOG.warning(f"File {fname} is empty of rows")
            data = pd.DataFrame()
        else:
            data = cls.read_csv_dataframe(fname, columns, header_pos)

        return cls(data, metadata, variable, meta_parser.units_map)

    @staticmethod
    def read_csv_dataframe(profile_filename, columns, header_position):
        """
        Read in a profile file. Managing the number of lines to skip and
        adjusting column names

        Args:
            profile_filename: Filename containing a manually measured
                             profile
            columns: list of columns to use in dataframe
            header_position: skiprows for pd.read_csv
        Returns:
            df: pd.dataframe contain csv data with desired column names
        """
        # header=0 because docs say to if using skip rows and columns
        df = pd.read_csv(
            profile_filename, header=0,
            skiprows=header_position,
            names=columns,
            encoding='latin'
        )
        # Special SMP specific tasks
        depth_fmt = 'snow_height'
        is_smp = False
        if 'force' in df.columns:
            # Convert depth from mm to cm
            df['depth'] = df['depth'].div(10)
            is_smp = True
            # Make the data negative from snow surface
            depth_fmt = 'surface_datum'

            # SMP serial number and original filename for provenance to the comment
            f = Path(profile_filename).name
            serial_no = f.split('SMP_')[-1][1:3]

            df['comments'] = f"fname = {f}, " \
                             f"serial no. = {serial_no}"

        if not df.empty:
            # Standardize all depth data
            new_depth = standardize_depth(
                df['depth'], desired_format=depth_fmt, is_smp=is_smp
            )

            if 'bottom_depth' in df.columns:
                delta = df['depth'] - new_depth
                df['bottom_depth'] = df['bottom_depth'] - delta

            df['depth'] = new_depth

            delta = abs(df['depth'].max() - df['depth'].min())
            LOG.debug(
                f'File contains a profile with'
                f' with {len(df)} layers across {delta:0.2f} cm'
            )
        if "flags" in df.columns:
            # Max length of the flags column
            df["flags"] = df["flags"].str.replace(" ", "")

        return df



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    "metloom>=0.4.0, <1.0",
    "utm",
    "numpy<2.0"
]

test_requirements = ['pytest>=3', ]

setup(
    author="M3 Works LLC",
    author_email='info@m3works.io',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    description="Weave insitu data together",
    entry_points={
        'console_scripts': [
            'insitupy=insitupy.cli:main',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='insitupy',
    name='insitupy',
    packages=find_packages(include=['insitupy', 'insitupy.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/m3works/insitupy',
    version='0.2.0',
    zip_safe=False,
)



from datetime import timedelta

import logging
from typing import List

import pandas as pd
import pytz
import utm

from .strings import StringManager
from ..variables import BaseMetadataVariables, BasePrimaryVariables
from ..profiles.metadata import ProfileMetaData

LOG = logging.getLogger(__name__)


class MetaDataParser:
    """
    Base class for parsing metadata
    """
    OUT_TIMEZONE = "UTC"
    ID_NAMES = ["pitid", "pit_id"]
    SITE_NAME_NAMES = ["location", "site_name"]
    LAT_NAMES = ["lat", "latitude"]
    LON_NAMES = ["lon", "lon", "longitude"]
    UTM_EPSG_PREFIX = "269"
    NORTHERN_HEMISPHERE = True
    PRIMARY_VARIABLES_CLASS = BasePrimaryVariables
    METADATA_VARIABLE_CLASS = BaseMetadataVariables

    def __init__(
        self, fname, timezone, header_sep=",", allow_split_lines=False,
        allow_map_failures=False,
        _id=None, campaign_name=None, units_map=None
    ):
        """
        Args:
            fname: path to file
            timezone: string timezone
            header_sep: expected header separator
            allow_split_lines: Allow for split header lines that
                don't start with the expected header character. In this case
                the number of header lines will be the max line starting with
                the expected character, and lines that don't start with
                that character will be combined with the previous line
            allow_map_failures: if a mapping fails, warn us and use the
                original string (default False)
            id: optional pass in to override id in parse_id
            campaign_name: optional override for campaign name
            units_map = optional map of variable type to MeasurementDescription
        """
        self._fname = fname
        self._input_timezone = timezone
        self._header_sep = header_sep
        self._rough_obj = {}
        self._lat_lon_easting_northing = None
        self._id = _id
        self._campaign_name = campaign_name
        self._units_map = units_map or {}

        self._allow_split_header_lines = allow_split_lines
        self._allow_map_failures = allow_map_failures

    @property
    def rough_obj(self):
        return self._rough_obj

    @property
    def units_map(self):
        return self._units_map

    @property
    def lat_lon_easting_northing(self):
        if self._lat_lon_easting_northing is None:
            self._lat_lon_easting_northing = self._parse_location()
        return self._lat_lon_easting_northing

    def parse_id(self) -> str:
        if self._id is not None:
            return self._id
        else:
            for k, v in self.rough_obj.items():
                if k in self.ID_NAMES:
                    return v

        raise RuntimeError(f"Failed to parse ID from {self.rough_obj}")

    def _handle_separate_datetime(self, keys, out_tz):
        """
        Handle a separate date and time entry

        Args:
            keys: list of keys
            out_tz: desired timezone
        Returns:
            parsed datetime
        """
        # Handle data dates and times
        if 'date' in keys and 'time' in keys:
            # Assume MMDDYY format
            if len(self.rough_obj['date']) == 6:
                dt = self.rough_obj['date']
                # Put into YY-MM-DD
                self.rough_obj['date'] = f'20{dt[-2:]}-{dt[0:2]}-{dt[2:4]}'
                # Allow for nan time
                self.rough_obj['time'] = StringManager.parse_none(
                    self.rough_obj['time']
                )

            date_str = self.rough_obj["date"]
            if self.rough_obj["time"] is not None:
                date_str += f" {self.rough_obj['time']}"
            d = pd.to_datetime(date_str)

        elif 'date' in keys:
            d = pd.to_datetime(self.rough_obj['date'])

        # Handle gpr data dates
        elif 'utcyear' in keys and 'utcdoy' in keys and 'utctod' in keys:
            base = pd.to_datetime(
                '{:d}-01-01 00:00:00 '.format(int(self.rough_obj['utcyear'])),
                utc=True)

            # Number of days since january 1
            d = int(self.rough_obj['utcdoy']) - 1

            # Zulu time (time without colons)
            time = str(self.rough_obj['utctod'])
            hr = int(time[0:2])  # hours
            mm = int(time[2:4])  # minutes
            ss = int(time[4:6])  # seconds
            ms = int(
                float('0.' + time.split('.')[-1]) * 1000)  # milliseconds

            delta = timedelta(
                days=d, hours=hr, minutes=mm, seconds=ss, milliseconds=ms
            )
            # This is the only key set that ignores in_timezone
            d = base.astimezone(pytz.timezone('UTC')) + delta
            d = d.astimezone(out_tz)

        else:
            raise ValueError(
                f'Data is missing date/time info!\n{self.rough_obj}'
            )
        return d

    def parse_date_time(self) -> pd.Timestamp:
        keys = [k.lower() for k in self.rough_obj.keys()]
        d = None
        out_tz = pytz.timezone(self.OUT_TIMEZONE)
        # Convert timezones if it is provided
        # this variable gets rewritten later
        in_timezone = self._input_timezone
        if in_timezone is not None:
            in_tz = pytz.timezone(in_timezone)
        # Otherwise assume incoming data is the same timezone
        # TODO: how do we handle row based timezone
        else:
            raise ValueError("We did not recieve a valid in_timezone")

        # Look for a single header entry containing date and time.
        for k in keys:
            kl = k.lower()
            if 'date' in kl and 'time' in kl:
                str_date = str(self.rough_obj[k].replace('T', '-'))
                d = pd.to_datetime(str_date)
                break

        # If we didn't find date/time combined.
        if d is None:
            d = self._handle_separate_datetime(keys, out_tz)

        if in_timezone is not None:
            d = d.tz_localize(in_tz)
            d = d.astimezone(out_tz)

        else:
            d.replace(tzinfo=out_tz)

        self.rough_obj['date'] = d.date()

        # Don't add time to a time that was nan or none
        if 'time' not in self.rough_obj.keys():
            self.rough_obj['time'] = d.timetz()
        else:
            if self.rough_obj['time'] is not None:
                self.rough_obj['time'] = d.timetz()

        dt_str = self.rough_obj["date"].isoformat()
        if self.rough_obj.get("time"):
            dt_str += f"T{self.rough_obj['time'].isoformat()}"
        dt = pd.to_datetime(dt_str)

        return dt

    def _parse_location_from_raw(self):
        """
        Initial parse of lat, lon, easting, northing from
        the rough object
        """
        lat = None
        lon = None
        easting = None
        northing = None
        for k, v in self.rough_obj.items():
            if k in self.LAT_NAMES:
                lat = float(v)
            elif k in self.LON_NAMES:
                lon = float(v)
            elif k == "easting":
                easting = v
            elif k == "northing":
                northing = v

        return lat, lon, easting, northing

    def lat_lon_from_easting_northing(self, easting, northing):
        zone_number = self.parse_utm_epsg()
        if isinstance(zone_number, str):
            raise RuntimeError(
                f"{zone_number} should be an integer: {self._fname}"
            )
        # Get the last two digits
        zone_number = int(str(zone_number)[-2:])
        try:
            lat, lon = utm.to_latlon(
                float(easting), float(northing), int(zone_number),
                northern=self.NORTHERN_HEMISPHERE)
        except Exception as e:
            LOG.error(e)
            raise RuntimeError(f"Failed with {easting}, {northing}")

        return lat, lon

    def _parse_location(self):
        """
        Parse the lat and lon from the rough input object
        Also parse the easting and northing
        # UTM Zone,13N
        # Easting,329131
        # Northing,4310328
        # Latitude,38.92524
        # Longitude,-106.97112
        # Flags,

        returns lat, lon, easting, northing
        """
        lat, lon, easting, northing = self._parse_location_from_raw()

        # Do nothing first
        if lat and lon:
            LOG.info("Latitude and Longitude parsed from the file")
        elif easting and northing:
            lat, lon = self.lat_lon_from_easting_northing(easting, northing)
        else:
            raise ValueError(
                f"Could not parse location from {self.rough_obj}"
            )
        return lat, lon, easting, northing

    def parse_latitude(self) -> float:
        return self.lat_lon_easting_northing[0]

    def parse_longitude(self) -> float:
        return self.lat_lon_easting_northing[1]

    def parse_utm_epsg(self) -> int:
        info = self.rough_obj
        epsg = None
        if 'utm_zone' in info.keys():
            utm_zone = int(
                ''.join([c for c in info['utm_zone'] if c.isnumeric()]))
            epsg = int(f"{self.UTM_EPSG_PREFIX}{utm_zone}")
        elif 'epsg' in info.keys():
            epsg = info["epsg"]
        # TODO: row based utm?
        return epsg

    def parse_campaign_name(self) -> str:
        if self._campaign_name is not None:
            return self._campaign_name
        for k, v in self.rough_obj.items():
            if k in self.SITE_NAME_NAMES:
                return v

        raise RuntimeError(f"Failed to parse Site Name from {self.rough_obj}")

    def parse_flags(self):
        result = None
        for k, v in self.rough_obj.items():
            if k in ["flags"]:
                result = v
                break

        return result

    def parse_observers(self) -> List[str]:
        result = None
        for k, v in self.rough_obj.items():
            if k in ["observers"]:
                entry = v
                result = entry.split(", ")
                result = [r.strip() for r in result]
                break

        return result

    def _preparse_meta(self, meta_lines):
        """
        Organize the header lines into a dictionary with lower case keys
        """
        # Key value pairs are separate by some separator provided.
        data = {}

        # Collect key value pairs from the information above the column header
        for ln in meta_lines:
            d = ln.split(self._header_sep)

            # Key is always the first entry in comma sep list
            k = StringManager.standardize_key(d[0])

            # Avoid splitting on times
            if 'time' in k or 'date' in k:
                value = ':'.join(d[1:]).strip()
            else:
                value = ', '.join(d[1:])
                value = StringManager.clean_str(value)

            # cast the rough object key to a known key
            known_name, k_mapping = self.METADATA_VARIABLE_CLASS.from_mapping(
                k, allow_failure=self._allow_map_failures
            )

            # Assign non-empty strings to dictionary
            if k and value:
                data[known_name] = value.strip(
                    ' '
                ).replace('"', '').replace('  ', ' ')

            elif k and not value:
                data[known_name] = None
        return data

    def parse(self):
        """
        Parse the file and return a metadata object.
        We can override these methods as needed to parse the different
        metadata

        This populates self.rough_obj

        Returns:
            (metadata object, column list, position of header in file)
        """
        (meta_lines, columns,
         columns_map, header_position) = self.find_header_info(self._fname)
        self._rough_obj = self._preparse_meta(meta_lines)
        # Create a standard metadata object
        metadata = ProfileMetaData(
            site_name=self.parse_id(),
            date_time=self.parse_date_time(),
            latitude=self.parse_latitude(),
            longitude=self.parse_longitude(),
            utm_epsg=str(self.parse_utm_epsg()),
            campaign_name=self.parse_campaign_name(),
            flags=self.parse_flags(),
            observers=self.parse_observers()
        )

        return metadata, columns, columns_map, header_position

    def _parse_header(self, lines):
        # Key value pairs are separate by some separator provided.
        data = {}

        # Collect key value pairs from the information above the column header
        for ln in lines:
            d = ln.split(self._header_sep)

            # Key is always the first entry in comma sep list
            k = StringManager.standardize_key(d[0])

            # Avoid splitting on times
            if 'time' in k or 'date' in k:
                value = ':'.join(d[1:]).strip()
            else:
                value = ', '.join(d[1:])
                value = StringManager.clean_str(value)

            # Assign non empty strings to dictionary
            if k and value:
                data[k] = value.strip(' ').replace('"', '').replace('  ', ' ')

            elif k and not value:
                data[k] = None

        LOG.debug(
            'Discovered {} lines of valid header info.'
            ''.format(len(data.keys()))
        )
        return data

    def _parse_columns(self, str_line):
        """
        Parse the column names from the input line. This can include mapping
        """
        # Parse the columns header based on the size of the last line
        # Remove units
        # for c in ['()', '[]']:
        #     str_line = StringManager.strip_encapsulated(str_line, c)

        raw_cols = str_line.strip('#').split(',')
        # Clean the raw columns
        standard_cols = [StringManager.standardize_key(c) for c in raw_cols]
        # Infer units from the raw columns
        infered_units = [StringManager.infer_unit_from_key(c) for c in raw_cols]
        final_cols = []
        final_col_map = {}
        inferred_units_map = {}
        # Iterate through the columns and map to desired result
        for c, unit in zip(standard_cols, infered_units):
            mapped_col, col_map = self.PRIMARY_VARIABLES_CLASS.from_mapping(
                c, allow_failure=self._allow_map_failures
            )
            # Store the list of columns to use when reading in the
            # dataframe
            final_cols.append(mapped_col)
            # Store the map of column name to the known variable
            final_col_map = {**final_col_map, **col_map}
            # Store the map of column name to inferred unit
            inferred_units_map[col_map[mapped_col].code] = unit

        return final_cols, final_col_map, inferred_units_map

    def find_header_info(self, filename=None):
        """
        Read in all site details file for a pit If the filename has the word
        site in it then we read everything in the file. Otherwise, we use this
        to read all the site data up to the header of the profile.

        E.g. Read all commented data until we see a column descriptor.

        Args:
            filename: Path to a csv containing # leading lines with site details

        Returns:
            tuple: **data** - Dictionary containing site details
                   **columns** - List of clean column names
                   **header_pos** - Index of the columns header for skiprows in
                                    read_csv
       """
        filename = filename or self._fname
        filename = str(filename)
        with open(filename, encoding='latin') as fp:
            lines = fp.readlines()
            fp.close()

        # Site description files have no need for column lists
        if 'site' in filename.lower():
            LOG.info('Parsing site description header...')
            columns = None
            header_pos = None
            header_indicator = None
            columns_map = {}
            self._units_map = self._units_map

        # Find the column names and where it is in the file
        else:
            header_pos, header_indicator = self._find_header_position(lines)
            # identify columns, map columns, and units map
            columns, columns_map, units_map = self._parse_columns(
                lines[header_pos]
            )
            # Combine with user defined units map
            self._units_map = {**self._units_map, **units_map}
            LOG.debug(
                f'Column Data found to be {len(columns)} columns based on'
                f' Line {header_pos}'
            )
            # Only parse what we know if the header
            lines = lines[0:header_pos]

        final_lines = lines

        # Clean up the lines from line returns to grab header info
        final_lines = [ln.strip() for ln in final_lines]
        # Join all data and split on header separator
        # This handles combining split lines
        str_data = " ".join(final_lines).split('#')
        str_data = [ln.strip() for ln in str_data if ln]

        return str_data, columns, columns_map, header_pos

    def _iterative_header_pos_search(self, lines, n_columns, header_indicator):
        # Use these to monitor if a larger column count is found
        header_pos = 0
        for i, l in enumerate(lines):
            if i == 0:
                previous = StringManager.get_alpha_ratio(lines[i])
            else:
                previous = StringManager.get_alpha_ratio(lines[i - 1])

            if StringManager.line_is_header(
                l, expected_columns=n_columns,
                header_indicator=header_indicator,
                previous_alpha_ratio=previous
            ):
                header_pos = i

            if i > header_pos:
                break
        return header_pos

    def _find_header_position(self, lines):
        """
        A flexible method that attempts to find and standardize column names
        for csv data. Looks for a comma separated line with N entries == to the
        last line in the file. If an entry is found with more commas than the
        last line then we use that. This allows us to have data that doesn't
        have all the commas in the data (SSA typically missing the comma for
        veg unless it was notable)

        Assumptions:

        1. The last line in file is of representative csv data
        2. The header is the last column that has more chars than numbers

        Args:
            lines: Complete list of strings from the file

        Returns:
            header position
        """

        # Minimum column size should match the last line of data (Assumption
        # #2)
        n_columns = len(lines[-1].split(','))

        if lines[0][0] == '#':
            header_indicator = '#'
        else:
            header_indicator = None

        if self._allow_split_header_lines:
            if header_indicator is None:
                raise RuntimeError(
                    "Cannot allow split lines with no clear header indicator"
                )
            else:
                # header pos is max lines with
                # first character == header indicator
                header_indices = [
                    index for index, value in enumerate(lines)
                    if value[0] == header_indicator
                ]
                header_pos = max(header_indices)

        else:
            header_pos = self._iterative_header_pos_search(
                lines, n_columns, header_indicator
            )

        LOG.debug('Found end of header at line {}...'.format(header_pos))
        return header_pos, header_indicator



from insitupy.variables.base_primary_variables import BasePrimaryVariables
from insitupy.variables.base_variables import MeasurementDescription


class SnowExPrimaryVariables(BasePrimaryVariables):
    """
    Extend the primary variables to include more mappings
    for snowex
    """

    EQUIVALENT_DIAMETER = MeasurementDescription(
        'equivalent_diameter', "",
        ['deq'], auto_remap=True
    )

    INSTRUMENT = MeasurementDescription(
        'instrument', "Instrument of measurement",
        ['smp_serial_number', 'measurement_tool', 'instrument'], auto_remap=True
    )
